import typing
import urllib.parse


def is_absolute_url(url) -> bool:
    u = urllib.parse.urlsplit(url, scheme='http')
    return u.netloc != ''


def escape_query(query: str) -> str:
    name, value = query.split('=', maxsplit=1)
    return urllib.parse.quote(name) + '=' + urllib.parse.quote(value)


def concat_urls(
        endpoints: typing.Tuple[str, ...],
        url: str,
        queries: typing.Tuple[str, ...],
) -> typing.Tuple[str, ...]:
    u = urllib.parse.urlsplit(url, scheme='http')
    escaped_queries = list(escape_query(q) for q in queries)
    if u.query:
        escaped_queries.insert(0, u.query)
    query_string = '&'.join(escaped_queries)

    if is_absolute_url(url):
        components = (
            # scheme
            u.scheme,
            # hostname
            u.netloc or 'localhost',
            # path
            u.path.lstrip('/'),
            # query
            query_string,
            # fragment
            '',
        )
        return urllib.parse.urlunsplit(components),

    concated_urls = []
    for endpoint in endpoints:
        e = urllib.parse.urlsplit(endpoint, scheme='http')

        components = (
            # scheme
            e.scheme,
            # hostname
            e.netloc or 'localhost',
            # path
            e.path.rstrip('/') + '/' + u.path.lstrip('/'),
            # query
            query_string,
            # fragment
            '',
        )
        concated_urls.append(urllib.parse.urlunsplit(components))

    return tuple(concated_urls)



import typing
from dataclasses import dataclass
from . import config
import requests
import jsonrpcclient  # type: ignore
import jsonrpcclient.exceptions  # type: ignore
from jsonrpcclient.requests import Request  # type: ignore
from jsonrpcclient.exceptions import ReceivedErrorResponseError, ReceivedNon2xxResponseError as InvalidResponseError  # type: ignore


class ConnectionError(Exception):
    pass


@dataclass(frozen=True)
class Endpoint:
    urls: typing.Tuple[str, ...]
    headers: typing.Tuple[config.HttpHeader, ...]
    basic: typing.Optional[config.BasicAuth]

    def send(self, req):
        for url in self.urls:
            try:
                return jsonrpcclient.send(url, req)
            except requests.ConnectionError:
                # Maybe... hostname or port is incorrect.
                # Try to other urls.
                continue
        raise ConnectionError()



from __future__ import annotations
import os
import typing
import dataclasses
from dataclasses import dataclass
from dataclasses_json import dataclass_json  # type: ignore

CONFIG_FILE_LOCATIONS = (
    './.apicall.json',
    '$GIT/.apicall.json',
    '~/.apicall.json',
    '~/.config/apicall.json',
)

# List of default ports.
#  * django: 8000/tcp
#  * bottle: 8080/tcp
#  * rails:  3000/tcp
#  * rack:   9292/tcp
DEFAULT_ENDPOINTS = (
    'http://localhost:8000',
    'http://localhost:8080',
    'http://localhost:3000',
    'http://localhost:9292',
)


class ConfigNotFound(Exception):
    pass


class GitRepositoryNotFound(Exception):
    pass


@dataclass_json
@dataclass(frozen=True)
class HttpHeader:
    name: str
    value: str


@dataclass_json
@dataclass(frozen=True)
class BasicAuth:
    user: str
    password: str


@dataclass_json
@dataclass(frozen=True)
class Config:
    headers: typing.Tuple[HttpHeader, ...] = dataclasses.field(  # type: ignore
        default_factory=tuple)
    basic: typing.Optional[BasicAuth] = dataclasses.field(default=None)
    endpoints: typing.Tuple[str, ...] = dataclasses.field(
        default=DEFAULT_ENDPOINTS)

    def remove_headers(self, names: typing.Iterable[str]) -> Config:
        """ Remove http headers with names.

        NOTE: I recommended to see documentation of add_header().
        """
        names = set(name.lower() for name in names)
        cleaned_headers = tuple(
            filter(lambda h: h.name.lower() not in names, self.headers))
        return dataclasses.replace(self, headers=cleaned_headers)

    def add_header(self, header: HttpHeader) -> Config:
        """ Add a http header.

        Add a http header.  If config already contains the header with the same name, that
        headers will be removed.

        NOTE:
        HTTP 1.1 specification allows same name headers.  But I think duplicate http headers
        are BAD PRACTICE because it can causes issues with some apps.  So I decide that
        duplicate HTTP headers feature does not support.
        https://stackoverflow.com/questions/3241326/set-more-than-one-http-header-with-the-same-name
        """
        cleaned = self.remove_headers([header.name])
        return dataclasses.replace(
            cleaned, headers=cleaned.headers + (header, ))


class FileSearcher:
    def __init__(self,
                 locations: typing.Iterable = CONFIG_FILE_LOCATIONS,
                 base_dir='.'):
        self.locations = locations
        self.base_dir = base_dir

    def find_git_root(self, base: str) -> str:
        p = os.path.abspath(base)
        while True:
            git_dir = os.path.join(p, '.git')
            if os.path.exists(git_dir):
                return p

            parent = os.path.dirname(p)
            if parent == p:
                # p is the root directory.
                break
            p = parent
        raise GitRepositoryNotFound()

    def expand_path(self, path: str) -> str:
        return os.path.abspath(os.path.expanduser(path))

    def search(self) -> str:
        base = os.path.abspath(self.base_dir)

        for pattern in self.locations:
            # Expands file path pattern.
            if pattern.startswith('$GIT/'):
                try:
                    git_root = self.find_git_root(base)
                    filepath = pattern.replace('$GIT/', git_root)
                except GitRepositoryNotFound:
                    continue
            else:
                filepath = self.expand_path(pattern)

            abspath = os.path.join(base, filepath)
            if os.path.exists(abspath):
                return abspath

        # All path patterns are checked.  But file is not found...
        raise ConfigNotFound()

    def default_location(self) -> str:
        base = os.path.abspath(self.base_dir)

        # 1. Git repository root.
        try:
            git_root = self.find_git_root(base)
            return os.path.join(git_root, '.apicall.json')
        except GitRepositoryNotFound:
            pass

        # 2. Home directory.
        return os.path.expanduser('~/.apicall.json')


def default() -> typing.Tuple[str, Config]:
    fname = FileSearcher().default_location()
    obj = Config()
    return fname, obj


def load() -> typing.Tuple[str, Config]:
    fname = FileSearcher().search()
    with open(fname) as f:
        js = f.read()

    # Workaround for an issue that data type of Config.headers[*] is wrong.
    # Expected type is HttpHeader object, but actual type is dict object.
    # This workaround converts the data type from dict to HttpHeader.
    #
    # The cause of this problem is bug of dataclasses_json library...
    # TODO: Bug report to dataclasses_json.
    conf = Config.from_json(js)  # type: ignore
    fixed_conf = dataclasses.replace(
        conf, headers=tuple(HttpHeader(**h) for h in conf.headers))

    return fname, fixed_conf


def load_or_default() -> typing.Tuple[str, Config]:
    try:
        return load()
    except ConfigNotFound:
        return default()


def save(path: str, conf: Config):
    js = conf.to_json()  # type: ignore
    with open(path, 'w') as f:
        f.write(js)



import sys
import io
import contextlib
import argparse
import typing
import os
import abc
import dataclasses
from . import config
from . import restapi
from . import urlutils
from . import printutils
from .printutils import pprint
from . import jsonrpc
from tabulate import tabulate  # type: ignore

# ExitCodes are compatible with curl command.
ExitCode = typing.NewType('ExitCode', int)
ExitOk = ExitCode(0)
ExitFailedToInit = ExitCode(1)
ExitInvalidArgs = ExitFailedToInit
ExitFailedToConnect = ExitCode(7)
ExitInvalidResponse = ExitCode(22)
ExitErrorReponse = ExitCode(254)
ExitSubprocessError = ExitCode(255)


def parse_headers(
        ns: argparse.Namespace) -> typing.Tuple[config.HttpHeader, ...]:
    """ --header引数をパースして、HttpHeaderオブジェクトのタプルを返す """

    def worker():
        for header in ns.header or []:
            name, value = header.split(':', 2)
            yield config.HttpHeader(name=name.rstrip(), value=value.lstrip())

        if ns.accept is not None:
            yield config.HttpHeader('accept', ns.accept)

        if ns.content_type is not None:
            yield config.HttpHeader('content-type', ns.content_type)

    return tuple(worker())


class CommandArgs(typing.NamedTuple):
    ns: argparse.Namespace
    args: typing.List[str]
    conf: config.Config
    conf_file: str


class TopCommand(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def build(self) -> argparse.ArgumentParser:
        pass

    # @abc.abstractmethod
    def __call__(self, ca: CommandArgs) -> ExitCode:
        print(ca.ns.format_help())
        return ExitInvalidArgs


class SubCommand(metaclass=abc.ABCMeta):
    NAME = 'dummy'

    def add_to(self, sp: argparse._SubParsersAction):
        p = sp.add_parser(self.NAME)
        p.set_defaults(fn=self, format_help=p.format_help)
        self.build_in(p)

    @abc.abstractmethod
    def build_in(self, p: argparse.ArgumentParser):
        pass

    # @abc.abstractmethod
    def __call__(self, ca: CommandArgs) -> ExitCode:
        print(ca.ns.format_help())
        return ExitInvalidArgs


################################################################
# Sub commands
class AuthHeaderSet(SubCommand):
    NAME = 'set'

    def build_in(self, p: argparse.ArgumentParser):
        p.add_argument('name')
        p.add_argument('value')

    def __call__(self, ca: CommandArgs) -> ExitCode:
        new_conf = ca.conf.add_header(
            config.HttpHeader(
                name=ca.ns.name,
                value=ca.ns.value,
            ))
        config.save(ca.conf_file, new_conf)
        return ExitOk


class AuthHeaderUnset(SubCommand):
    NAME = 'unset'

    def build_in(self, p: argparse.ArgumentParser):
        p.add_argument('names', nargs='+')

    def __call__(self, ca: CommandArgs) -> ExitCode:
        new_conf = ca.conf.remove_headers(ca.ns.names)
        config.save(ca.conf_file, new_conf)
        return ExitOk


class AuthHeader(SubCommand):
    NAME = 'header'

    def build_in(self, p: argparse.ArgumentParser):
        sp = p.add_subparsers()
        AuthHeaderSet().add_to(sp)
        AuthHeaderUnset().add_to(sp)

    def __call__(self, ca: CommandArgs) -> ExitCode:
        data = tuple((h.name, h.value) for h in ca.conf.headers)
        tabule = tabulate(data, headers=('NAME', 'VALUE'))
        print(tabule)
        return ExitOk


class AuthBasicSet(SubCommand):
    NAME = 'set'

    def build_in(self, p: argparse.ArgumentParser):
        p.add_argument('user')
        p.add_argument('password')

    def __call__(self, ca: CommandArgs) -> ExitCode:
        new_conf = dataclasses.replace(
            ca.conf,
            basic=config.BasicAuth(
                user=ca.ns.user,
                password=ca.ns.password,
            ))
        config.save(ca.conf_file, new_conf)
        return ExitOk


class AuthBasicUnset(SubCommand):
    NAME = 'unset'

    def build_in(self, p: argparse.ArgumentParser):
        pass

    def __call__(self, ca: CommandArgs) -> ExitCode:
        new_conf = dataclasses.replace(ca.conf, basic=None)
        config.save(ca.conf_file, new_conf)
        return ExitOk


class AuthBasic(SubCommand):
    NAME = 'basic'

    def build_in(self, p: argparse.ArgumentParser):
        sp = p.add_subparsers()
        AuthBasicSet().add_to(sp)
        AuthBasicUnset().add_to(sp)

    def __call__(self, ca: CommandArgs) -> ExitCode:
        basic = ca.conf.basic
        if basic is None:
            print('Basic authentication is not configured.')
            return ExitOk
        else:
            print(f'User: {basic.user}')
            print(f'Password: {basic.password}')
            return ExitOk


class Auth(SubCommand):
    NAME = 'auth'

    def build_in(self, p: argparse.ArgumentParser):
        sp = p.add_subparsers()
        AuthHeader().add_to(sp)
        AuthBasic().add_to(sp)


class Endpoint(SubCommand):
    NAME = 'endpoint'

    def build_in(self, p: argparse.ArgumentParser):
        p.add_argument('urls', nargs='*')

    def __call__(self, ca: CommandArgs) -> ExitCode:
        urls = ca.ns.urls
        if urls:
            new_conf = dataclasses.replace(ca.conf, endpoints=tuple(urls))
            config.save(ca.conf_file, new_conf)
            return ExitOk
        else:
            print('\n'.join(ca.conf.endpoints))
            return ExitOk


class Rest(SubCommand):
    NAME = 'rest'

    def build_in(self, p: argparse.ArgumentParser):
        # Options are compatible with curl command.
        # That is better not to override to curl options.
        p.add_argument('-R', '--raw', action='store_true')
        p.add_argument('-v', '--verbose', action='count', default=0)
        p.add_argument('-H', '--header', action='append')
        p.add_argument('--accept')
        p.add_argument('--content-type', '--type')
        p.add_argument('-d', '--data')
        p.add_argument('method')
        p.add_argument('url')
        p.add_argument('queries', nargs='*')

    def __call__(self, ca: CommandArgs) -> ExitCode:
        # 正しいcontent-type headerをつけていないので、アプリケーション側で正しく処理してくれない。
        try:
            data = self.parse_data(ca.ns.data)
            response = restapi.Request(
                method=ca.ns.method,
                urls=urlutils.concat_urls(
                    endpoints=ca.conf.endpoints,
                    url=ca.ns.url,
                    queries=tuple(ca.ns.queries),
                ),
                headers=(self.detect_data_type(data) + ca.conf.headers +
                         parse_headers(ca.ns)),
                basic=ca.conf.basic,
                data=data,
            ).fetch(
                verbose=ca.ns.verbose,
                logging_cb=lambda msg: print(msg),
            )
        except restapi.ConnectionError:
            pprint(
                'ERROR: Could not connect to server.\n'
                'Please check endpoint urls and HTTP server status.',
                file=sys.stderr)
            return ExitFailedToConnect

        try:
            pprint(response.raw_body)
            return ExitOk
        except printutils.SubprocessError:
            return ExitSubprocessError

    def parse_data(self, data: typing.Optional[str]) -> typing.Optional[bytes]:
        if data is None:
            return None
        elif data.startswith('@'):
            fname = data[1:]
            with open(fname, 'rb') as f:
                return f.read()
        else:
            return bytes(data, 'utf8')

    def detect_data_type(self, data: typing.Optional[bytes]
                         ) -> typing.Tuple[config.HttpHeader, ...]:
        """ --data引数で指定したデータのタイプを判別し、適切なcontent-typeヘッダーを返す  """
        if data is None:
            return tuple()
        if printutils.is_unicode_string(data):
            if printutils.is_json(data):
                return config.HttpHeader('content-type', 'application/json'),
            return config.HttpHeader('content-type', 'text/plain'),
        return config.HttpHeader('content-type', 'application/octet-stream'),


class Jsonrpc(SubCommand):
    NAME = 'jsonrpc'

    def build_in(self, p: argparse.ArgumentParser):
        p.add_argument('--raw', action='store_true')
        p.add_argument('--verbose', action='count', default=0)
        p.add_argument('-H', '--header', action='append')
        p.add_argument('--accept')
        p.add_argument('--content-type', '--type')
        p.add_argument('method')
        p.add_argument('args', nargs='*')

    def __call__(self, ca: CommandArgs) -> ExitCode:
        try:
            res = jsonrpc.Endpoint(
                urls=ca.conf.endpoints,
                headers=ca.conf.headers + parse_headers(ca.ns),
                basic=ca.conf.basic,
            ).send(jsonrpc.Request(
                ca.ns.method,
                ca.ns.args,
            ))
        except jsonrpc.ConnectionError:
            print(
                'ERROR: Could not connect to server.\n'
                'Please check endpoint urls and HTTP server status.',
                file=sys.stderr)
            return ExitFailedToConnect
        except jsonrpc.InvalidResponseError as e:
            pprint(str(e).encode())
            return ExitInvalidResponse
        except jsonrpc.ReceivedErrorResponseError as e:
            pprint(str(e.response).encode())
            return ExitErrorReponse

        try:
            pprint(res.text.encode())
            return ExitOk
        except printutils.SubprocessError:
            return ExitSubprocessError


################################################################
# Top level commands
class ApicallCommand(TopCommand):
    def build(self) -> argparse.ArgumentParser:
        p = argparse.ArgumentParser(prog='apicall')
        p.set_defaults(fn=self, format_help=p.format_help)
        sp = p.add_subparsers()
        Auth().add_to(sp)
        Endpoint().add_to(sp)
        Rest().add_to(sp)
        Jsonrpc().add_to(sp)
        return p


class RestcallCommand(TopCommand):
    def build(self) -> argparse.ArgumentParser:
        p = argparse.ArgumentParser(prog='restcall')

        rest = Rest()
        p.set_defaults(fn=rest, format_help=p.format_help)
        rest.build_in(p)
        return p


class JsonrpccallCommand(TopCommand):
    def build(self) -> argparse.ArgumentParser:
        p = argparse.ArgumentParser(prog='jsonrpccall')

        jr = Jsonrpc()
        p.set_defaults(fn=jr, format_help=p.format_help)
        jr.build_in(p)
        return p


################################################################
# Parser
class ParseResult(typing.NamedTuple):
    success: bool
    ns: typing.Optional[argparse.Namespace]
    output: str


def parse(prog, args) -> ParseResult:
    basename = os.path.basename(prog)

    cmd: TopCommand
    if basename == 'restcall':
        cmd = RestcallCommand()
    elif basename == 'jsonrpccall':
        cmd = JsonrpccallCommand()
    else:
        cmd = ApicallCommand()

    ns: typing.Optional[argparse.Namespace] = None
    output = io.StringIO()
    exit_code = None
    try:
        # Redirect stdout/stderr to on-memory buffer.
        with contextlib.redirect_stdout(output), \
             contextlib.redirect_stderr(output):
            ns = cmd.build().parse_args(args)
            success = True
    except SystemExit:
        success = False

    return ParseResult(
        success=success,
        ns=ns,
        output=output.getvalue(),
    )



import typing
from dataclasses import dataclass
from . import config
import requests

SHOW_HEADERS = 1


class ConnectionError(Exception):
    pass


@dataclass(frozen=True)
class Request:
    method: str
    urls: typing.Tuple[str, ...]
    headers: typing.Tuple[config.HttpHeader, ...]
    basic: typing.Optional[config.BasicAuth]
    data: typing.Optional[bytes]

    @property
    def dict_headers(self):
        return {h.name: h.value for h in self.headers}

    def fetch(
            self,
            verbose: int = 0,
            logging_cb: typing.Optional[typing.Callable[[str], None]] = None):
        for url in self.urls:
            try:
                res = requests.request(
                    self.method,
                    url,
                    headers=self.dict_headers,
                    data=self.data,
                )

                self.logging_request(verbose, logging_cb, res)

                # The request is successful.  Return the response object.
                return Response(
                    request=self,
                    _result=res,
                )
            except requests.ConnectionError:
                # Maybe... hostname or port is incorrect.
                # Try to other urls.
                continue

        # Tried all urls, but I could not connect them all.
        raise ConnectionError()

    def logging_request(self, verbose: int,
                        cb: typing.Optional[typing.Callable[[str], None]],
                        res: requests.Response):
        if cb is None:
            return
        if verbose < SHOW_HEADERS:
            return

        first = True
        for r in res.history + [res]:  # type: requests.Response
            if not first:
                cb('')
            first = False

            cb(f'> {r.request.method} {r.request.url}')
            for header, value in r.request.headers.items():
                cb(f'> {header}: {value}')
            cb('')
            cb(f'< {r.status_code} {r.reason}')
            for header, value in r.headers.items():
                cb(f'< {header}: {value}')


@dataclass(frozen=True)
class Response:
    request: Request
    _result: requests.Response

    @property
    def body(self) -> str:
        return self._result.text

    @property
    def raw_body(self) -> bytes:
        return self._result.content

    @property
    def turnaround_time(self):
        return self._result.elapsed

    @property
    def content_type(self) -> typing.Optional[str]:
        if 'content-type' not in self._result.headers:
            return None

        h = self._result.headers['content-type']
        if ';' in h:
            # Strip MIME directives.
            h, _ = h.split(';', 1)
        return h



# Suppress warnings
import warnings
warnings.filterwarnings(
    'ignore',
    message='^`NoneType` object ',
    module='^dataclasses_json\.',
    category=RuntimeWarning)

import sys

from . import arguments
from . import config


def main():
    result = arguments.parse(sys.argv[0], sys.argv[1:])
    if not result.success:
        # Arguments is not valid.
        # Show error messages and exit.
        sys.stderr.write(result.output)
        return arguments.ExitInvalidArgs

    conf_file, conf = config.load_or_default()
    ca = arguments.CommandArgs(
        args=sys.argv,
        ns=result.ns,
        conf=conf,
        conf_file=conf_file,
    )
    return result.ns.fn(ca)



import sys
import tempfile
import subprocess
import typing
import json
import contextlib


class SubprocessError(Exception):
    pass


def _run(args, stdin):
    """ Run the process.

    contextlibでstdout/stderrをリダイレクトしている状況下で
    サブプロセスを起動した場合、サブプロセスの出力はリダイレクト前の
    fdに出力される。
    この関数は、この問題へのワークアラウンドを提供する。 """
    return subprocess.run(
        args, stdin=stdin, stdout=sys.stdout, stderr=sys.stderr)


def is_unicode_string(data: bytes) -> bool:
    try:
        data.decode()
        return True
    except UnicodeDecodeError:
        return False


def is_json(data: bytes) -> bool:
    try:
        json.loads(data.decode())
        return True
    except json.JSONDecodeError:
        return False


def print_hex(data: bytes):
    with tempfile.TemporaryFile() as f:
        f.write(data)
        f.seek(0)
        result = _run(['od', '-t', 'x1'], stdin=f)
        if result.returncode != 0:
            raise SubprocessError()


def print_json(data: bytes):
    """ Print JSON to terminal through jq command. """
    with tempfile.TemporaryFile() as f:
        f.write(data)
        f.seek(0)
        result = _run(['jq'], stdin=f)
        if result.returncode != 0:
            raise SubprocessError()


def pprint(data: typing.Union[bytes, str], raw: bool = False, file=None):
    # convert data to bytes.
    if isinstance(data, str):
        data = data.encode()

    context = contextlib.nullcontext()  # type: ignore
    if file is not None:
        context = contextlib.redirect_stdout(file)

    with context:
        if raw or not sys.stdout.isatty():
            sys.stdout.buffer.write(data)
            return

        if is_unicode_string(data):
            if is_json(data):
                print_json(data)
                return
            sys.stdout.buffer.write(data)
            return

        # Attempt to write binary data to terminal !
        # Show warning messages and writes data in hex representation.
        sys.stderr.write(
            'WARNING: The response body is displayed in hex representation\n'
            '         because it is binary data.\n')
        print_hex(data)
